import os
import pygsheets
import pandas as pd
from pygsheets import DataRange

from .creds import load_creds
load_creds()

class Gsheets:
    """
    Pass in a service account file path and query data stored in Google Sheets
    Example: Gsheets('SPREADSHEET').query('WORKSHEET')
    """
    def __init__(self, spreadsheet):
        self.client = pygsheets.authorize(service_file=(os.getenv("GSHEETS_SERVICE_ACCOUNT")))
        self.spreadsheet_key = os.getenv(spreadsheet)
        self.spreadsheet = self.client.open_by_key(self.spreadsheet_key)

    def query(self, worksheet, start_cell=None, end_cell=None):
        """
        Returns results of spreadsheet, worksheet as a pandas DataFrame
        Spreadsheet is stored as env var
        Worksheet is passed in as plain text
        """        
        ws = self.spreadsheet.worksheet_by_title(worksheet)
        df = ws.get_as_df(start=start_cell, end=end_cell)
        print(f"{self.spreadsheet.title}.{worksheet}: Downloaded {df.shape} shape")
        return df

    # If get_as_df does not work, use this instead
    def select_all_values(self, worksheet):
        return pd.DataFrame(
            self.spreadsheet.worksheet_by_title(worksheet).get_all_records()
        )

    def write_df(self, df, worksheet, start_cell='A1'):
        self.spreadsheet.worksheet_by_title(worksheet).set_dataframe(df, start=start_cell)

    def write_df_with_timestamp(self, df, worksheet):
        # Reference the worksheet by title
        worksheet = self.spreadsheet.worksheet_by_title(worksheet)
        # Clear existing data
        worksheet.clear(start='A1')
        # Write the frame to the worksheet, adding rows if necessary
        worksheet.set_dataframe(df, start="A1", extend=True)
        # Resize the worksheet to the shape of the frame
        worksheet.resize(rows=df.shape[0]+1, cols=df.shape[1])
        # Add borders
        drange = DataRange(start='A1', worksheet=worksheet)
        drange.update_borders(top=True, right=True, bottom=True, left=True, inner_horizontal=True, inner_vertical=True, style='SOLID')
        # Complete message
        print(f"Wrote {len(df)} rows to {self.spreadsheet.title}.{worksheet.title}")


import os
import json
import pendulum
import requests


class Slack:

    """
    Pass in workspace as a Slack token to send 
    messages to channels using plain requests
    """

    def __init__(self, workspace):
        self.TOKEN = os.getenv(workspace)
        self.div = "─────────────────────"

    @staticmethod
    def slackify(message):
        return f'```{message}```'

    def send_message(self, channel, message, blocks=None):
        message += f'\n{self.div}'

        response = requests.post("https://slack.com/api/chat.postMessage",
                             data={'token': self.TOKEN,
                                   'channel': channel,
                                   'text': message,
                                   'blocks': json.dumps(blocks) if blocks else None}).json()

        if not response['ok']:
            print(f"Invalid response. Error: {response['error']}")
        else:
            return response

    def send_file(self, channels, comment, filename):

        response = requests.post("https://slack.com/api/files.upload",
                             data={'token': self.TOKEN,
                                   'channels': channels,
                                   'initial_comment': comment},
                             files={'file': open(filename, 'rb')}).json()
        self.send_message(channel=channels, message='')

        if not response['ok']:
            print(f"Invalid response. Error: {response['error']}")
        else:
            return response

    def notify_yourself(self, message):

        now = pendulum.now(tz="Asia/Singapore").strftime('%Y-%m-%d %H:%M:%S %p')

        self.send_message(
            channel="UCFNX4DQ9",
            message=f"{now} - {message}"
        )


import os
import pendulum
import pandas as pd

# BigQuery API
from google.cloud import bigquery
from google.oauth2 import service_account
from google.cloud.exceptions import NotFound
from .transform import Transform


class BigQuery:

    """
    Pass in a service account file path and query data stored in BigQuery
    Example: BigQuery().query("SELECT * FROM `jinlee.dbt.fct_receive` LIMIT 10")
    """
    
    def __init__(self):
        credentials = service_account.Credentials.from_service_account_file(
            os.getenv("BIGQUERY_SERVICE_ACCOUNT"), scopes=["https://www.googleapis.com/auth/cloud-platform"],
        )
        self.client = bigquery.Client(credentials=credentials, project=credentials.project_id)
        self.project = self.client.project

    def query(self, sql):
        """
        Returns results of SQL Query as a pandas DataFrame
        """
        return self.client.query(sql).to_dataframe()

    @staticmethod
    def prepare_df(df):
        # Lower case columns, replace spaces with underscore
        df.columns = df.columns.str.lower().str.replace(' ', '_')
        
        # Add a UTC timestamp
        df['updated_at'] = pendulum.now().strftime('%Y-%m-%d %H:%M:%S %p')
        df['updated_at'] = df['updated_at'].apply(pd.to_datetime, format='%Y-%m-%d %H:%M:%S %p', utc=True)
        
        return df

    def list_datasets(self):
    
        """
        Datasets in project jinlee:
            accounts
            dbo
            dbt_bigquery
            dev
            gsheets
            jinlee
            spareparts
        """
    
        datasets = list(self.client.list_datasets())  # Make an API request.
        project = self.client.project

        if datasets:
            
            dataset_list = []
            
            print("Datasets in project {}:".format(project))
            for dataset in datasets:
                print("\t{}".format(dataset.dataset_id))
                dataset_list.append(dataset.dataset_id)
                
            return dataset_list
            
        else:
            print("{} project does not contain any datasets.".format(project))
    
    def list_tables_in_dataset(self, dataset):
        
        # TODO(developer): Set dataset_id to the ID of the dataset that contains
        #                  the tables you are listing.
        # dataset_id = 'jinlee.dbo'

        tables = self.client.list_tables(f"{self.project}.{dataset}")  # Make an API request.
        table_names = []
        print("Tables contained in '{}':".format(dataset))
        for table in tables:
            table_name = "{}.{}.{}".format(table.project, table.dataset_id, table.table_id)
            table_names.append(table_name)
            print(table_name)
        return table_names
    
    # READ
    
    def select(self, sql, query_config=None):
        
        """
        Pass in a SQL Query and query_config
        and return the results as a DataFrame        
        """

        # self.client.query(sql).to_dataframe()
        
#         sql = """
#             SELECT * 
#             FROM `jinlee.dbo.weighbridge_PMF_MASTER`
#             WHERE DATE(date) >= @date
#             LIMIT @limit
#         """

#         query_config = bigquery.QueryJobConfig(
#             query_parameters=[
#                 bigquery.ScalarQueryParameter('date', 'DATE', '2021-01-01'),
#                 bigquery.ScalarQueryParameter('limit', 'INTEGER', 1000)
#             ]
#         )
    
        return self.client.query(sql, job_config=query_config).to_dataframe()
    
    # CREATE

    def updated_count(self, table_id):

        """
        Checks how many times a script has run
        """

        update_counter = 0

        # Query distinct count and add to counter
        query_job = self.client.query(
            f"""
            SELECT count(distinct(updated_at)) as updated_count
            FROM {table_id}
            """
        )

        print(f"[ Read  ] Checking if {table_id} has been updated before ...")

        # Query will only run if table is found
        try:
            results = query_job.result() 
            for row in results:
                update_counter += row.updated_count
            print(f"[ Read  ] Table {table_id} has been updated {update_counter} time(s).")

        except NotFound:
            print(f"[ Read  ] Table {table_id} does not exist. This will be the first time writing to BQ.")

        return update_counter

    def check_if_dataset_exists(self, dataset_id):

        """
        Checks if the dataset {project_name}.{schema} exists
        Create it if it does not
        """

        try:
            self.client.get_dataset(dataset_id) # Make an API request
            print(f"[ Read  ] Dataset {dataset_id} found! Proceeding to write data ...")

        except NotFound:
            print(f"[ Read  ] {dataset_id} not found, attempting to create the requested dataset now ...")
            dataset = bigquery.Dataset(dataset_id)
            dataset.location = "asia-southeast1"
            dataset = self.client.create_dataset(dataset, timeout=30)
            print(f"[ Write ] {self.client.project}.{dataset.dataset_id} created in {dataset.location}.")

    def create_table(self, df, source, schema, table, write_behavior=None):

        # table_id = 'jinlee.stocks'

        # Since string columns use the "object" dtype, pass in a (partial) schema
        # to ensure the correct BigQuery data type.
        # job_config = bigquery.LoadJobConfig(schema=[
        #     bigquery.SchemaField("my_string", "STRING"),
        # ])

        """
        `WRITE_TRUNCATE`: If the table already exists, BigQuery overwrites the table data and uses the schema from the query result.
        `WRITE_APPEND`: If the table already exists, BigQuery appends the data to the table.
        `WRITE_EMPTY`: If the table already exists and contains data, a 'duplicate' error is returned in the job result.

        The default value is `WRITE_EMPTY`. Each action is atomic and only occurs if BigQuery is able to complete the job successfully. 
        Creation, truncation and append actions occur as one atomic update upon job completion.
        """
        
        dataset_id = f"{self.project}.{schema}"
        table_id = f"{dataset_id}.{table}"

        # Default behavior is append
        if write_behavior is None:
            write_behavior = "WRITE_APPEND"

        # Check for empty frame
        if df is not None:

            self.check_if_dataset_exists(dataset_id=dataset_id)
            
            # Identify where the data came from
            df['data_source'] = source

            # Prepare the df to be written to BigQuery
            transformed_df = Transform(df).lower_case_columns()\
                                          .add_updated_timestamp().df

            # Keep track of how many times the data has been inserted
            df['updated_count'] = self.updated_count(table_id=table_id)                                          

            # Proceed to write the data to the dataset
            job_config = bigquery.LoadJobConfig(
                write_disposition=write_behavior
            )
            
            print(f"[ Write ] BigQuery.{table_id} using {write_behavior}.")
            
            job = self.client.load_table_from_dataframe(
                df, table_id, job_config=job_config
            )
            
            print(f"[ Done  ] BigQuery.{table_id} Wrote {len(transformed_df)} rows.")

            # Wait for the load job to complete.
            return job.result()

        else:
            print(f"No data was written to BigQuery - Empty data.")
        
    # DELETE
    
    def delete_dataset(self, dataset_id):
        
        # TODO(developer): Set model_id to the ID of the model to fetch.
        # Example: dataset_id = 'jinlee'

        # Use the delete_contents parameter to delete a dataset and its contents.
        # Use the not_found_ok parameter to not receive an error if the dataset has already been deleted.
        self.client.delete_dataset(
            dataset_id, delete_contents=True, not_found_ok=True
        )  # Make an API request.

        print("Deleted dataset '{}'.".format(dataset_id))
        
    # DELETE TABLE
    
    def delete_table(self, table_id):
        
        # TODO(developer): Set table_id to the ID of the table to fetch.
        # table_id = 'your-project.your_dataset.your_table'

        # If the table does not exist, delete_table raises
        # google.api_core.exceptions.NotFound unless not_found_ok is True.
        self.client.delete_table(table_id, not_found_ok=True)  # Make an API request.
        
        print("Deleted table '{}'.".format(table_id))                


import os
import pickle
import pandas as pd
from googleapiclient.discovery import build

from .creds import load_creds
load_creds()

with open(os.getenv("GOOGLE_DRIVE_PICKLE"), 'rb') as f:
    creds = pickle.load(f)

class Drive:
    """
    Using the built service, this Class
    enables you to move files around from
    one folder to another
    """
    def __init__(self, source, target):
        self.service = build(
            "drive", 'v3',
            credentials=creds
        )
        self.source = source
        self.target = target

    def fetch_source_files(self):

        query = f"parents = '{self.source}'"
        response = self.service.files().list(q=query).execute()
        files = response.get('files')
        nextPageToken = response.get('nextPageToken')

        # A unique token is generated for each page view of length 100
        while nextPageToken:
            response = self.service.files().list(
                q=query,
                pageToken=nextPageToken).execute()

            files.extend(response.get('files'))
            nextPageToken = response.get('nextPageToken')

        print(f"Found {len(files)} files in source folder SQL Backup.")

        return files

    def move_source_to_target(self):

        """
        Only keep the last 10 days of backups
        in the SQL Backup folder and move the 
        rest to Long Term (so that our server disk
        which is running the Google Drive client will
        not be full)
        """

        files = self.fetch_source_files()

        df = pd.DataFrame(files)

        # Date filtering
        df['date'] = df['name'].str.split('_').str[0].apply(pd.to_datetime, format='%Y-%m-%d')
        ten_days_ago = (pd.to_datetime('today') - pd.Timedelta('10 days')).strftime('%Y-%m-%d')

        files_to_move = df[df['date'] < ten_days_ago].to_dict('records')
        files_remaining = df[df['date'] >= ten_days_ago]

        min_date = files_remaining['date'].min().strftime('%Y-%m-%d')
        max_date = files_remaining['date'].max().strftime('%Y-%m-%d')

        # Move only the required files

        success = 0

        for index, file in enumerate(files_to_move, 1):
            if file['mimeType'] != 'application/vnd.google-apps.folder':
                print(
                    f"[File {index} of {len(files_to_move)}] Moving {file['name']} to SQL Backup (Long Term) ...")

                try:
                    self.service.files().update(
                        fileId=file.get('id'),
                        addParents=self.target,
                        removeParents=self.source
                    ).execute()

                    success += 1

                except Exception as e:
                    print(f"Moving the file {file['name']} failed.")
                    print(e)

        print(f"Successfully moved {success} out of {len(files)} files.")
        print(f"There are {len(files_remaining)} backup files remaining in SQL Backup.")
        print(f"We kept the last 10 days of backup: {min_date} - {max_date}.")


import pendulum
import pandas as pd


class Transform:
    
    def __init__(self, df):
        self.df = df

    def lower_case_columns(self):
        self.df.columns = self.df.columns.str.lower().str.replace(' ', '_')
        return self

    def add_updated_timestamp(self):
        # Add a UTC timestamp
        self.df['updated_at'] = pendulum.now().strftime('%Y-%m-%d %H:%M:%S %p')
        self.df['updated_at'] = self.df['updated_at'].apply(pd.to_datetime, format='%Y-%m-%d %H:%M:%S %p', utc=True)
        return self

    def filter_today(self, date_fmt=None):

        """
        Pass in a pandas dataframe to:

        1. Lower case col names
        2. Cast date col as datetime
        3. Set datetime col as index
        4. Return where datetime == today
        """

        if date_fmt is None:
            date_fmt = '%d/%m/%Y'

        today = pendulum.today().naive()

        self.df.columns = self.df.columns.str.lower()
        self.df['date'] = pd.to_datetime(self.df['date'], format=date_fmt)
        self.df.set_index('date', inplace=True)
        return self.df.loc[today]

    def filter_up_to_today(self, date_fmt=None):

        if date_fmt is None:
            date_fmt = '%d/%m/%Y'

        today = pendulum.today().naive()

        self.df.columns = self.df.columns.str.lower()
        self.df['date'] = pd.to_datetime(self.df['date'], format=date_fmt)
        self.df.set_index('date', inplace=True)
        return self.df.loc[:today]


import os
import pendulum
import pandas as pd
from sqlalchemy import create_engine


class SQLServer:
    """
    Pass in a conn string to generate the engine for querying
    Example: pd.read_sql(query, SQLServer(SERVER="EXAMPLE_SERVER_CONN_STRING").engine)
    
    SQLAlchemy EXAMPLE_SERVER_CONN_STRING Format:
    mssql+pyodbc://user:password@server-ip/database?driver=ODBC+Driver
    """
    def __init__(self, SQLSERVER_CONN_STR):
        self.engine = create_engine(os.getenv(SQLSERVER_CONN_STR))
        self.database = self.engine.url.database
        self.dialect = str(self.engine.url).split('+')[0]
        self.timestamp = pendulum.now(tz='Asia/Singapore').strftime('%Y-%m-%d %H:%M:%S %p')

    def read_with_timestamp(self, sql):
        df = pd.read_sql(sql, self.engine)
        df["updated_at"] = pd.to_datetime('now', utc=True).tz_convert('Asia/Singapore').strftime('%Y-%m-%d %H:%M:%S %p')
        return df


import os
import boto3
from botocore.exceptions import ClientError

from .creds import load_creds
load_creds()

class S3:
    # Pass in your bucket name. Defaults to flowstate
    def __init__(self, bucket=None):
        self.client = boto3.client(
            's3',
            aws_access_key_id = os.getenv("AWS-ACCESS-KEY-ID")
            aws_secret_access_key = os.os.getenv("AWS-SECRET-ACCESS-KEY")
        )
        self.bucket = "flowstate" if bucket is None else bucket

    def list_buckets(self):
        """
        Listing buckets
        :return:
        """
        for key in self.client.list_buckets()['Buckets']:
            print(key['Name'])

    def list_objects_in_bucket(self, folder=None):

        folder = '' if folder is None else folder

        for key in self.client.list_objects(
            Bucket=self.bucket,
            Prefix=folder)['Contents']:

            print(key['Key'])

    def delete(self, key):
        self.client.delete_object(Bucket=self.bucket, Key=key)

    def upload(self, file_name, object_name=None, **extra_args):
        """Upload a file to an S3 bucket

        :param file_name: File to upload
        :param object_name: S3 object name. If not specified then file_name is used
        :param extra_args: Pass in ExtraArgs={'ACL': 'public-read'} for public access
        :return: True if file was uploaded, else False

        -- USAGE -- 
        
        from src.utils.s3 import S3
        FILE_NAME = 'google-drive-oauth-2.0.json'
        S3().upload(file_name=FILE_NAME, object_name=f'creds/{FILE_NAME}')

        """

        # If S3 object_name was not specified, use file_name
        if object_name is None:
            object_name = file_name

        try:
            response = self.client.upload_file(
                file_name, self.bucket, object_name,
                **extra_args
            )

            print(f"Upload success: https://s3-ap-southeast-1.amazonaws.com/{self.bucket}/{object_name}")

        except ClientError as e:
            logger.error(e)
            return False
        
        return response



    def upload_csv(self, df, folder_name, file_name, **extra_args):
        """
        Converts a DataFrame to csv file and uploads it to S3
        """
        file_name = f"{file_name}.csv"
        df.to_csv(file_name, index=False)
        self.upload(file_name=f'./{file_name}', object_name=f"{folder_name}/{file_name}", **extra_args)
        os.remove(f'./{file_name}')




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['data_connectors']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.37,<2.0.0',
 'db-dtypes>=1.0.2,<2.0.0',
 'google-api-python-client>=2.51.0,<3.0.0',
 'google-auth-httplib2>=0.1.0,<0.2.0',
 'google-auth-oauthlib>=0.5.2,<0.6.0',
 'google-cloud-bigquery>=3.2.0,<4.0.0',
 'pandas>=1.4.2,<2.0.0',
 'pendulum>=2.1.2,<3.0.0',
 'pygsheets>=2.0.5,<3.0.0',
 'pyodbc>=4.0.32,<5.0.0',
 'python-dotenv>=0.19.1,<0.20.0',
 'streamlit-aggrid>=0.2.3,<0.3.0']

setup_kwargs = {
    'name': 'data-connectors',
    'version': '0.20.20',
    'description': 'Connect to various data sources - SQL Server, Google Sheets, Google BigQuery',
    'long_description': None,
    'author': 'CA Lee',
    'author_email': 'flowstate_crypto@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from os import path
from dotenv import load_dotenv
from pathlib import Path

def load_creds():
    ENV_FILE_PATH = '/root/.creds/.env'

    if not path.exists(ENV_FILE_PATH):
        ENV_FILE_PATH = input("Please enter the absolute path to your .env file on your local file system.")

    dotenv_path = Path(ENV_FILE_PATH)
    load_dotenv(dotenv_path=dotenv_path)


from st_aggrid import AgGrid
from st_aggrid.shared import GridUpdateMode
from st_aggrid.grid_options_builder import GridOptionsBuilder

class Grid:
    def __init__(self, df):
        self.df = df
        self.gb = GridOptionsBuilder.from_dataframe(df)
        self.gb.configure_pagination()
        self.gb.configure_side_bar()
        self.gb.configure_default_column(groupable=True, value=True, enableRowGroup=True, aggFunc="sum", editable=True)
        self.gridOptions = self.gb.build()
        
    def render(self, theme_select, grid_height=500):

        # theme : str, optional
        #         theme used by ag-grid. One of:
        #             'streamlit' -> follows default streamlit colors
        #             'light'     -> ag-grid balham-light theme
        #             'dark'      -> ag-grid balham-dark theme
        #             'blue'      -> ag-grid blue theme
        #             'fresh'     -> ag-grid fresh theme
        #             'material'  -> ag-grid material theme
        #         By default 'light'        

        return AgGrid(
            self.df, 
            height = grid_height, 
            gridOptions=self.gridOptions, 
            enable_enterprise_modules=True, 
            fit_columns_on_grid_load=True, 
            theme=theme_select,
            update_mode=GridUpdateMode.SELECTION_CHANGED
        )